"""Get data from Belvis rest api."""

from __future__ import annotations

import datetime as dt
import json
import pathlib
import urllib
from dataclasses import dataclass
from typing import Callable, Dict, List, Union

import numpy as np
import pandas as pd
import requests
import yaml

from .common import print_status


def create_url(server: str, path: str, *queryparts: str) -> str:
    """Convenience function to build URL for querying a server."""
    sp = f"{server}{path}"
    query = "&".join([urllib.parse.quote(qp, safe=":=") for qp in queryparts])
    url = sp if not query else f"{sp}?{query}"
    return url


@dataclass
class Access:
    # Zero-parameter function to authorize with server.
    authenticate: Callable[[]]
    # Function used to fetch data from the server.
    request: Callable[[str], requests.request]

    def __post_init__(self):
        self.authenticate()  # Run once to verify authentication works.

    @classmethod
    def from_usr_pwd(cls, server: str, tenant: str, usr: str, pwd: str) -> Access:
        """Use username and password to access Belvis API.

        Parameters
        ----------
        server : str
            Address of Belvis server, incl port.
        tenant : str
            Name of Belvis tenant.
        usr : str
            Username.
        pwd : str
            Password.
        """
        session = requests.Session()
        auth_url = create_url(
            server, "/rest/session", f"tenant={tenant}", f"usr={usr}", f"pwd={pwd}"
        )

        def auth() -> None:
            response = session.get(auth_url)
            if response.status_code != 200:
                raise PermissionError(
                    f"Please check authentication details: {response}"
                )

        def req(url: str) -> requests.request:
            return session.get(url)

        return cls(auth, req)


class Api:
    """A class to interact with a Belvis Rest API server."""

    @classmethod
    def from_file(cls, filepath: Union[str, pathlib.Path]) -> Api:
        """Load api data from file.

        Parameters
        ----------
        filepath : Union[str, pathlib.Path]
            Path to load yaml api configuration from. Also the path to save api configuration
            (which includes the cache).

        Notes
        -----
        The api configuration file must have the keys 'server' (= string of server address
        including port), 'tenant' (= name of the belvis tenant), and possibly 'cache'
        (= dictionary).
        """
        api_yml = yaml.load(open(filepath), Loader=yaml.FullLoader)
        return cls(**api_yml, filepath=filepath)

    def __init__(
        self,
        server: str,
        tenant: str,
        cache: Dict[str, Dict[str, int]] = None,
        filepath: pathlib.Path = None,
    ) -> None:
        """
        Parameters
        ----------
        server : str
            Server address including port.
        tenant : str
            Name of the Belvis tenant.
        cache : Dict[str, Dict[str, int]], optional
            Cache to use for lookup. Nested dictionany with format {pfid: {tsname: tsid}}
        filepath : pathlib.Path, optional
            If specified, saves api data (server, tenant, cache) to this file after every
            change for persistent storage. If not specified, the data is cached only for
            the duration of the class instance.
        """
        self._retry = True
        self._server = server
        self._tenant = tenant
        self._cache = cache if cache is not None else {}
        self._filepath = filepath
        # Must be set later.
        self._access: Access = None

    server: str = property(lambda self: self._server)
    tenant: str = property(lambda self: self._tenant)
    cache: Dict = property(lambda self: self._cache)
    filepath: pathlib.Path = property(lambda self: self._filepath)

    @filepath.setter
    def filepath(self, path: pathlib.Path):
        self._filepath = path

    @property
    def access(self) -> Access:
        if self._access is None:
            raise ValueError(
                "No access method has been set yet. Set with .access_from_usr_pwd()."
            )
        return self._access

    # ---

    def access_from_usr_pwd(self, usr: str, pwd: str) -> None:
        """Specify username and password to access Belvis API.

        Parameters
        ----------
        usr : str
            Username.
        pwd : str
            Password.
        """
        self._access = Access.from_usr_pwd(self.server, self.tenant, usr, pwd)

    def to_file(self, filepath: pathlib.Path) -> None:
        """Write api configuration to file. Access method is not stored.

        Parameters
        ----------
        filepath : pathlib.Path
            Path of file to store data to.
        """
        yaml.dump(
            {"server": self.server, "tenant": self.tenant, "cache": self.cache},
            open(filepath, "w"),
            allow_unicode=True,
            sort_keys=False,
        )

    def read_cache(self, pfid: str, tsname: str) -> int:
        """Retrieve belvis timeseries id from cache file. Returns None if not found.
        Case sensitive.

        Parameters
        ----------
        pfid : str
            Id (= short name) of portfolio.
        tsname : str
            Name of timeseries.

        Returns
        -------
        int
        """
        return self.cache.get(pfid, {}).get(tsname)

    def write_cache(self, pfid: str, tsname_to_tsid: Dict[str, int]) -> None:
        """Write belvis timeseries id into cache file. Overwrites if already present.

        Parameters
        ----------
        pfid : str
            Id (= short name) of the Belvis portfolio.
        tsname_to_tsid : Dict[str, int]
            Dictionary mapping timeseries names to their ids.
        """
        # Insert new data.
        if pfid not in self.cache:
            self.cache[pfid] = {}
        self.cache[pfid].update(tsname_to_tsid)
        # Save file.
        if self.filepath is not None:
            self.to_file(self.filepath)

    def query(self, url: str) -> Union[Dict, List]:
        """Query connection for general information."""
        try:
            response = self.access.request(url)

        except ConnectionError as e:
            raise ConnectionError("Check connection to server.") from e

        if response.status_code == 200:
            # Success.
            self._retry = True
            return json.loads(response.text)
        elif self._retry is True:
            # Authentication might have expired, or maybe has never been done before. Try once.
            self._retry = False
            self.access.authenticate()
            return self.query(url)  # retry.
        else:
            raise RuntimeError(response)

    def metadata(self, tsid: int) -> Dict:
        """Get information about timeseries.

        Parameters
        ----------
        tsid : int
            Belvis id of timeseries.

        Returns
        -------
        dict
            Metadata about the timeseries.
        """
        url = create_url(
            self.server, f"/rest/energy/belvis/{self.tenant}/timeSeries/{tsid}"
        )
        return self.query(url)

    def all_ts(self, pfid: str) -> Dict[str, int]:
        """Use API to find all timeseries in a Belvis portfolio and store in cache. May
        take a long time if portfolio has many timeseries.

        Parameters
        ----------
        pfid : str
            Id (= short name) of Belvis portfolio.

        Returns
        -------
        Dict[str, int]
            Dictionary with found timeseries. Keys are the timeseries names, the values are
            the timeseries ids.
        """
        # Get all timeseries ids in the portfolio.
        url = create_url(
            self.server,
            f"/rest/energy/belvis/{self.tenant}/timeSeries",
            f"instancetoken={pfid}",
        )
        paths = self.query(url)
        tsids = [int(path.split("/")[-1]) for path in paths]

        # Get name of each timeseries id.
        found = {}
        for tsid in tsids:
            info = self.metadata(tsid)
            tsid, tsname = int(info["id"]), info["timeSeriesName"]
            found[tsname] = tsid

        # Save ALL that are found to cache.
        self.write_cache(pfid, found)
        return found

    def find_tsids(
        self,
        pfid: str,
        tsname: str = "",
        case_sensitive: bool = True,
        allow_partial: bool = True,
    ) -> Dict[str, int]:
        """Use API to find specific timeseries in a Belvis portfolio. May take a long
        time if portfolio has many timeseries.

        Parameters
        ----------
        pfid : str
            Id (= short name) of Belvis portfolio.
        tsname : str, optional (default: return all timeseries).
            Name of timeseries.
        case_sensitive : bool, optional (default: True)
            If True, do a case sensitive search for the timeseries name.
        allow_partial : bool, optional (default: True)
            If True, also returns timeseries if the name partially matches.

        Returns
        -------
        Dict[str, int]
            Dictionary with found timeseries. Keys are the timeseries names, the values are
            the timeseries ids.
        """
        if not case_sensitive:
            tsname = tsname.lower()

        def matchingname(found: str) -> bool:
            if not case_sensitive:
                found = found.lower()
            if allow_partial:
                return tsname in found
            else:
                return tsname == found

        # Filter on timeseries name.
        all_ts = self.all_ts(pfid)
        return {tsname: tsid for tsname, tsid in all_ts.items() if matchingname(tsname)}

    def find_tsid(self, pfid: str, tsname: str) -> int:
        """Find Belvis id number of timeseries.

        Parameters
        ----------
        pfid : str
            Id (=short name) of portfolio as found in Belvis.
        tsname : str
            Full name of timeseries.

        Returns
        -------
        int
            id of found timeseries.
        """
        # Try to get from cache first.
        if (tsid := self.read_cache(pfid, tsname)) is not None:
            return tsid

        # No cached value found: get from API.
        print_status(
            f"Timeseries '{tsname}' in portfolio '{pfid}' not found. If cache is complete and"
            " up to date, this should not happen, and ``pfid`` or ``tsname`` may be incorrect."
        )
        hits = self.find_tsids(pfid, tsname, allow_partial=False)
        # Raise error if 0 or > 1 found.
        if len(hits) == 0:
            raise ValueError(
                f"No timeseries with exact name '{tsname}' found in portfolio '{pfid}'."
                " Check if ``pfid`` and ``tsname`` are correct."
            )
        elif len(hits) > 1:
            raise ValueError(
                f"Multiple timeseries with exact name '{tsname}' found in portfolio '{pfid}'."
                " Check your Belvis instance."
            )
        return next(iter(hits.values()))

    def series(
        self,
        tsid: int,
        ts_left: Union[pd.Timestamp, dt.datetime],
        ts_right: Union[pd.Timestamp, dt.datetime],
        *,
        leftrange: str = "exclusive",
        rightrange: str = "inclusive",
        missing2zero: bool = True,
        blocking: bool = True,
    ) -> pd.Series:
        """Return timeseries in given delivery time interval using its id.

        Parameters
        ----------
        tsid : int
            Belvis id of timeseries.
        ts_left : Union[pd.Timestamp, dt.datetime]
        ts_right : Union[pd.Timestamp, dt.datetime]
        leftrange : str, optional (default: 'exclusive')
            'inclusive' ('exclusive') to get values with timestamp that is >= (>) ts_left.
            Default: 'exclusive' because timestamps in Belvis are *usually* right-bound.
        rightrange : str, optional (default: 'inclusive')
            'inclusive' ('exclusive') to get values with timestamp that is <= (<) ts_right.
            Default: 'inclusive' because timestamps in Belvis are *usually* right-bound.
        missing2zero : bool, optional (default: True)
            What to do with values that are flagged as 'missing'. True to replace with 0,
            False to replace with nan.
        blocking : bool, optional (default: True)
            If True, recalculates data that is not up-to-date before returning; might take
            long time or result in internal-server-error. If False, return most up-to-date
            data that is available without recalculating.

        Returns
        -------
        pd.Series
            with resulting information.

        Notes
        -----
        Returns series with data as found in Belvis; no correction (e.g. for right-bounded
        timestamps) done.
        """
        # Get metadata and print status message.
        metadata = self.metadata(tsid)
        # Get data.
        url = create_url(
            self.server,
            f"/rest/energy/belvis/{self.tenant}/timeSeries/{tsid}/values",
            f"timeRange={ts_left.isoformat()}--{ts_right.isoformat()}",
            f"timeRangeType={leftrange}-{rightrange}",
            f"blocking={str(blocking).lower()}",
        )
        records = self.query(url)
        df = pd.DataFrame.from_records(records)
        # Replace missing values.
        mask = df["pf"] == "missing"
        df.loc[mask, "v"] = 0.0 if missing2zero else np.nan
        # Find unit.
        unit = metadata["measurementUnit"]
        # Turn into series.
        # (int ->) float -> pint to suppress conversion warning
        values, index = df["v"].to_list(), pd.DatetimeIndex(df["ts"])
        s = pd.Series(values, index, float).astype(f"pint[{unit}]")
        s.index.freq = pd.infer_freq(s.index)
        return s

    def series_from_tsname(
        self,
        pfid: str,
        tsname: str,
        ts_left: Union[pd.Timestamp, dt.datetime],
        ts_right: Union[pd.Timestamp, dt.datetime],
        *,
        leftrange: str = "exclusive",
        rightrange: str = "inclusive",
        missing2zero: bool = True,
        blocking: bool = True,
    ) -> pd.Series:
        """Return timeseries in given delivery time interval, using its name and portfolio
        id.

        Parameters
        ----------
        pfid : int
            ID (=short name) of portfolio in Belvis.
        tsname : str
            Name of the timeseries. Must be exact.
        ts_left : Union[pd.Timestamp, dt.datetime]
        ts_right : Union[pd.Timestamp, dt.datetime]
        leftrange : str, optional (default: 'exclusive')
            'inclusive' ('exclusive') to get values with timestamp that is >= (>) ts_left.
            Default: 'exclusive' because timestamps in Belvis are *usually* right-bound.
        rightrange : str, optional (default: 'inclusive')
            'inclusive' ('exclusive') to get values with timestamp that is <= (<) ts_right.
            Default: 'inclusive' because timestamps in Belvis are *usually* right-bound.
        missing2zero : bool, optional (default: True)
            What to do with values that are flagged as 'missing'. True to replace with 0,
            False to replace with nan.
        blocking : bool, optional (default: True)
            If True, recalculates data that is not up-to-date before returning; might take
            long time or result in internal-server-error. If False, return most up-to-date
            data that is available without recalculating.

        Returns
        -------
        pd.Series
            with resulting information.

        Notes
        -----
        - Returns series with data as found in Belvis; no correction (e.g. for right-bounded
          timestamps) done.
        - If not yet cached, the ``.series()`` method is potentially a lot faster.
        """
        tsid = self.find_tsid(pfid, tsname)
        return self.series(
            tsid,
            ts_left,
            ts_right,
            leftrange=leftrange,
            rightrange=rightrange,
            missing2zero=missing2zero,
            blocking=blocking,
        )



"""Getting data from Kisters' portfolio management system 'Belvis'."""

from . import _version, adjustment
from .api import Api
from .example import (
    example_api,
    example_api_to_file,
    example_structure,
    example_structure_to_file,
)
from .structure import Structure, Ts, TsTree
from .tenant import Tenant

__version__ = _version.get_versions()["version"]



from __future__ import annotations

import pathlib
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, Union

import pandas as pd
import yaml

TsNameTree = Union[str, Iterable[str], Dict[str, "TsNameTree"]]


@dataclass
class Ts:
    pfid: str
    name: str
    tsid: int = None
    series: pd.Series = None


TsTree = Union[Ts, Iterable[Ts], Dict[str, "TsTree"]]


def create_tstree(pfid: str, tsname_tree: TsNameTree) -> TsTree:
    if isinstance(tsname_tree, Dict):
        return {k: create_tstree(pfid, subtree) for k, subtree in tsname_tree.items()}
    elif isinstance(tsname_tree, str):
        tsname = tsname_tree
        return [Ts(pfid, tsname)]
    elif isinstance(tsname_tree, Iterable):
        tsnames = tsname_tree
        return [Ts(pfid, tsname) for tsname in tsnames]


@dataclass
class Portfolios:
    original: Iterable[str]
    synthetic: Dict[str, Union[str, Iterable[str]]] = field(default_factory=dict)

    def __post_init__(self):
        # Assert that the class doesn't have missing refences.
        if not isinstance(self.original, Iterable):
            raise TypeError(f"'original' must be Iterable, got {type(self.original)}.")
        if not len(self.original):
            raise ValueError("'original' must have at least one element.")
        for ref in self.synthetic.values():
            try:
                self._assert_valid_ref(ref)
            except AssertionError as e:
                raise ValueError("Unexpected value for 'synthetic'") from e

    def _assert_valid_ref(self, ref: str) -> None:
        if isinstance(ref, str):
            assert ref in self.original or ref in self.synthetic.keys()
        elif isinstance(ref, Iterable):
            for r in ref:
                self._assert_valid_ref(r)
        else:
            raise AssertionError(
                f"Incorrect type for ``ref``: expected str or Iterable; got {type(ref)}."
            )


@dataclass
class Structure:
    """Structure of the belvis database; which portfolios and timeseries are of interest.

    Parameters
    ----------
    freq : str
        Frequency of the spot market, which is also the shortest frequency that is of
        interest, e.g. quarterhourly ('15T') for power and daily ('D') for gas. Ideally,
        all data in the Belvis database is available at this frequency.
    tz : str
        The timezone of the spot market and the Belvis database. It is the timezone that
        the data is converted into.
    pfline : Dict[str, TsNameTree]
        The portfolio lines that we are interested in.
        Dictionary with string keys (=pfline ids), which map onto the following possible
        values:

        - string value or list of string values (=timeseries name(s) as found in Belvis),
          which define a flat portfolio line (i.e., without children).
        - dictionary, which defines a mapping of child names onto timeseries name(s). May
          be nested.

    portfolios : Portfolios
        The portfolios we are interested in.
        Dictionary with at least the key 'original' which maps onto a list of strings (=
        the ids (=short names) of portfolios as found in Belvis). May also have a key
        'synthetic' which maps onto a dictionary, of which the keys are user-defined
        portfolio ids and the values are the original pfids that must be added.
    prices : Dict[str, Dict[str, Any]]
        The prices we are interested in.
        Dictionary with string keys (=price ids), which map onto a dictionary that has
        the keys 'pfid' (which has the id of the portfolio as its value) and 'tsnames'
        (which has a (list of) timeseries names as its value).
    corrections : Dict[str, Dict[str, Union[TsNameTree, None]]]
        If not all pflines can be found in all portfolios, we can specify this here.

        - If a portfolio has a pfline that is not found in the others (and therefore not
          specified in parameter ``pfline`` above), we can add a nested dictionary
          {pfid: {pflineid: tsnames}}.
        - If a porfolio does not have a pfline that is found in the others (and therefore
          specified in parameter ``pfline`` above), we can exclude it by adding the
          nested dictionary {pfid: {pflineid: None}}.

    Notes
    -----
    May be loaded from yaml file with the .from_file() class method.
    """

    freq: str
    tz: str
    pflines: Dict[str, TsNameTree]
    portfolios: Portfolios
    prices: Dict[str, Dict[str, Any]]
    corrections: Dict[str, Dict[str, Union[TsNameTree, None]]] = field(
        default_factory=dict
    )

    def _expand_tree(self, tsname_tree: TsNameTree) -> TsNameTree:
        """Check if tree refers to existing value, and replace if yes."""
        if tsname_tree is None:
            return tsname_tree
        elif isinstance(tsname_tree, Dict):
            return {k: self._expand_tree(ts) for k, ts in tsname_tree.items()}
        elif isinstance(tsname_tree, str):
            if tsname_tree in self.pflines:
                return self._expand_tree(self.pflines[tsname_tree])
            else:
                return tsname_tree
        else:
            return [self._expand_tree(ts) for ts in tsname_tree]

    def __post_init__(self):
        # Finish initialisation.
        self.portfolios = Portfolios(**self.portfolios)

        # Ensure pflines part is correct.
        self.pflines = {
            pflineid: self._expand_tree(tsname_tree)
            for pflineid, tsname_tree in self.pflines.items()
        }

        # Assert corrections part is valid.
        for pfid, pflinesdict in self.corrections.items():
            assert pfid in self.portfolios.original  # must override existing portfolio
            for pfline, ref in pflinesdict.items():
                if ref is None:
                    assert pfline in self.pflines.keys()

        # Assert prices part is valid.
        for pricedict in self.prices.values():
            assert "pfid" in pricedict
            assert "tsnames" in pricedict
            assert (
                isinstance(pricedict["tsnames"], str)
                or isinstance(pricedict["tsnames"], Dict)
                or isinstance(pricedict["tsnames"], Iterable)
            )

    @classmethod
    def from_file(cls, filepath: Union[str, pathlib.Path]) -> Structure:
        """Load structure from file.

        Parameters
        ----------
        filepath : Union[str, pathlib.Path]
            Path to load yaml structure file from.
        """
        conf_yml = yaml.load(open(filepath), Loader=yaml.FullLoader)
        return cls(**conf_yml)

    def tstree_pfline(self, pfid: str, pflineid: str) -> TsTree:
        """Timeseries that must be fetched to get the wanted pflineid for a given pfid.

        Parameters
        ----------
        pfid : str
            Id of portfolio. Must be original.
        pflineid : str
            Id of portfolio line.

        Returns
        -------
        TsTree
            Names of timeseries that must be fetched.
        """
        if pfid not in (avail := self.available_pfids(True)):
            raise ValueError(f"``pfid`` must one of {', '.join(avail)}.")

        # Use correction, if it exists.
        tsname_tree = self.corrections.get(pfid, {}).get(pflineid, "notfound")
        if tsname_tree is None:
            raise ValueError(
                f"The portfolio line ({pflineid}) does not exists for this portfolio ({pfid})."
            )
        elif tsname_tree != "notfound":
            return create_tstree(pfid, tsname_tree)

        # If not, use default.
        tsname_tree = self.pflines.get(pflineid, None)
        if tsname_tree is None:
            raise ValueError(
                f"The portfolio line ({pflineid}) does not exists for this portfolio ({pfid})."
            )
        return create_tstree(pfid, tsname_tree)

    def tstree_price(self, priceid: str) -> TsTree:
        """Timeseries that must be fetched to get the wanted price.

        Parameters
        ----------
        priceid : str
            Id of price.

        Returns
        -------
        TsTree
            Timeseries that must be fetched.
        """
        if priceid not in (avail := self.available_priceids()):
            raise ValueError(f"``priceid`` must be one of {', '.join(avail)}.")

        pfid = self.prices[priceid]["pfid"]
        tsname_tree = self.prices[priceid]["tsnames"]
        return create_tstree(pfid, tsname_tree)

    def available_pfids(self, original_only: bool = False) -> Iterable[str]:
        """All available portfolio ids.

        Parameters
        ----------
        original_only : bool, optional (default: False)
            If True, only return the ("original") portfolio ids that are present in belvis,
            and omit the ("synthetic") ones that are created by summing.

        Returns
        -------
        Iterable[str]
            List of portfolio ids that exist in this tenant, according to the configuration.
        """
        if original_only:
            return self.portfolios.original
        else:
            return [*self.portfolios.original, *self.portfolios.synthetic.keys()]

    def to_original_pfids(self, pfid: str) -> Iterable[str]:
        """The original pfids that need to be added to get values of wanted pfid.

        Parameters
        ----------
        pfid : str
            Id of portfolio. May be original or synthetic.

        Returns
        -------
        Iterable[str]
            List of original portfolio ids.
        """
        if pfid in self.portfolios.original:
            return [pfid]
        elif pfid in self.portfolios.synthetic:
            pfids = []
            for child_pfid in self.portfolios.synthetic[pfid]:
                for pfid in self.to_original_pfids(child_pfid):
                    pfids.append(pfid)
            return pfids
        raise ValueError(
            f"Could not find portfolio id {pfid} in configuration for this tenant; "
            f"expected one of {', '.join((*self.portfolios.original, *self.portfolios.synthetic))}."
        )

    def available_pflineids(self, pfid: str) -> Iterable[str]:
        """All available portfolio line ids for a given portfolio.

        Parameters
        ----------
        pfid : str
            Id of portfolio. May be original or synthetic.

        Returns
        -------
        Iterable[str]
            Set of portfolio line ids that exist in this tenant and portfolio, according
            to the configuration.
        """
        if pfid in self.portfolios.original:
            # Default.
            pflineids = set(self.pflines.keys())
            # Corrections.
            for pflineid, tsnames in self.corrections.get(pfid, {}).items():
                if tsnames is None:
                    pflineids.remove(pflineid)
                else:
                    pflineids.add(pflineid)
            # Return.
            return pflineids
        else:
            # Only keep those pflineids that are present in all summands.
            original_pfids = self.to_original_pfids(pfid)
            return list(
                set.intersection(
                    *[self.available_pflineids(pfid) for pfid in original_pfids]
                )
            )

    def available_priceids(self) -> Iterable[str]:
        """All available price ids for this tenant.

        Parameters
        ----------
        None

        Returns
        -------
        Iterable[str]
            List of price ids that exist in this tenant, according to the configuration.
        """
        return self.prices.keys()



import pathlib
import shutil
from typing import Union

from .api import Api
from .structure import Structure


def _example_filepath(obj: str, which: str) -> pathlib.Path:
    """Location of example file."""
    filename = f"example_{obj}_{which}.yaml"
    path = pathlib.Path(__file__).parent / filename
    if not path.exists():
        raise FileNotFoundError(f"No file found at this path: {path}")
    return path


def _checked_filepath(filepath: Union[str, pathlib.Path]) -> pathlib.Path:
    """Correct and check filepath."""
    if isinstance(filepath, str):
        filepath = pathlib.Path(filepath)
    if not filepath.isfile():
        raise ValueError("Must specify path to a file.")
    elif filepath.exists():
        raise ValueError("File already exists; delete first.")
    return filepath


def example_structure(which: str = "basic") -> Structure:
    """Example instance of Structure class.

    Parameters
    ----------
    which : {'basic' (default), 'complex'}
        Which example to load.

    Returns
    -------
    Structure
    """
    return Structure.from_file(_example_filepath("structure", which))


def example_api(which: str = "basic") -> Api:
    """Example instance of Api class.

    Parameters
    ----------
    which : {'basic' (default), 'complex'}
        Which example to load.

    Returns
    -------
    Api
    """
    return Api.from_file(_example_filepath("api", which))


def example_structure_to_file(
    filepath: Union[str, pathlib.Path], which: str = "basic"
) -> None:
    """Save an example of a structure configuration to file.

    Parameters
    ----------
    path : Union[str, pathlib.Path]
        Path to save the file to.
    which : {'basic' (default), 'complex'}
        Which example to save.
    """
    shutil.copy(_example_filepath("structure", which), _checked_filepath(filepath))


def example_api_to_file(
    filepath: Union[str, pathlib.Path], which: str = "basic"
) -> None:
    """Save an example of an api configuration to file.

    Parameters
    ----------
    path : Union[str, pathlib.Path]
        Path to save the file to.
    which : {'basic' (default), 'complex'}
        Which example to save.
    """
    shutil.copy(_example_filepath("api", which), _checked_filepath(filepath))



from collections import defaultdict
from typing import Dict, Iterable

import pandas as pd
import portfolyo as pf

from . import adjustment
from .adjustment import Aftercare
from .api import Api
from .common import print_status
from .structure import Structure, Ts, TsTree


def fact_default_aftercare(tz) -> Aftercare:

    convert_to_tz = adjustment.fact_convert_to_tz(tz)

    def aftercare(s: pd.Series, tsid: int, pfid: str, tsname: str) -> pd.Series:
        s = convert_to_tz(s)
        s = adjustment.infer_frequency(s)
        s = adjustment.makeleft(s)
        return s

    return aftercare


class Tenant:
    """Class to interact with belvis tenants and retrieve data.

    Notes
    -----
    An important attribute of this class is .aftercare, which is a function that is used
    to make small adjustments to the timeseries returned by the .api.series() method.
    This may be necessary due to local Belvis settings. The function must accept 4
    arguments: the pandas Series, the timeseries id, the portfolio id, and the timeseries
    name. It must return a pandas Series from which a portfolio line (portfolyo.PfLine)
    can be initialized.

    - By default, the function makes three adjustments: it sets the timezone (to the one
      specified in .structure.tz), it infers the frequency of the timeseries, and it makes
      the timestamps leftbound. (by assuming the timestamps returned by Belvis are
      leftbound if they represent daily (or longer) time periods, and rightbound
      otherwise.
    - If the default is not what is needed, the user must specify a custom aftercare
      function. If a specific timeseries needs a special treatment, the function parameters
      (the timeseries id, the pfid, the timeseries name) can be used in the function
      definition to target that timeseries only. If they are not needed, ``*args`` may
      be used to "eat up" the unneeded parameters.
    - There are several ready-made adjustment functions available in the ``belvys.adjustment``
      module.

    If .portfolio_pfl() and .price_pfl() raise an Exception when creating the ``PfLine``
    instance, inspect the series that the PfLine receives as input, by setting their
    ``debug`` parameter to True.
    """

    def __init__(self, structure: Structure = None, api: Api = None):
        """
        Parameters
        ----------
        structure : Structure, optional
            The structure of the objects found in this tenant.
        api : Api, optional
            The connection to the rest api for this tenant.
        """
        self._structure = structure
        self._api = api
        self._aftercare = fact_default_aftercare(self.structure.tz)

    @property
    def structure(self) -> Structure:
        if self._structure is None:
            raise ValueError("No structure has been set yet. Set with .structure = ...")
        return self._structure

    @structure.setter
    def structure(self, structure: structure) -> None:
        self._structure = structure

    @property
    def api(self) -> Api:
        if self._api is None:
            raise ValueError("No API object has been set yet. Set with .api = ...")
        return self._api

    @api.setter
    def api(self, api: Api) -> None:
        self._api = api

    @property
    def aftercare(self) -> Aftercare:
        return self._aftercare

    @aftercare.setter
    def aftercare(self, aftercare: Aftercare) -> None:
        self._aftercare = aftercare

    # ---

    def update_cache(self) -> None:
        """Loop through all portfolios that may be fetched, find the timeseries they contain,
        and store all in cache. May take long time!
        """
        pfids_1 = self.structure.available_pfids(True)
        pfids_2 = [price["pfid"] for price in self.structure.prices.values()]
        for pfid in set([*pfids_1, *pfids_2]):
            _ = self.api.all_ts(pfid)

    def _fetch_series(self, ts_tree: TsTree, ts_left, ts_right, **kwargs) -> None:
        if isinstance(ts_tree, Ts):
            ts = ts_tree
            if ts.tsid is None:
                ts.tsid = self.api.find_tsid(ts.pfid, ts.name)
            print_status(f"{ts.pfid} | {ts.tsid} | {ts.name} | [{ts_left}-{ts_right}]")
            ts.series = self.api.series(ts.tsid, ts_left, ts_right, **kwargs)
        if isinstance(ts_tree, Dict):
            for subtree in ts_tree.values():
                self._fetch_series(subtree, ts_left, ts_right, **kwargs)
        elif isinstance(ts_tree, Iterable):  # Must be after Dict
            for branch in ts_tree:
                self._fetch_series(branch, ts_left, ts_right, **kwargs)

    def _pfline(self, ts_tree: TsTree) -> pf.PfLine:
        """Create a portfolio line from the data stored in the ts_tree series."""

        # MultiPfLine.

        if isinstance(ts_tree, Dict):
            children = {}
            for name, subtree in ts_tree.items():
                children[name] = self._pfline(subtree)
            return pf.PfLine(children)

        # SinglePfLine.

        tss = [ts_tree] if isinstance(ts_tree, Ts) else ts_tree
        # Collect the timeseries values from Belvis.
        series_by_dimension = defaultdict(list)
        for ts in tss:
            s = ts.series  # series as returned by api.
            # Here we make corrections for design decisions in our Belvis database.
            if self.aftercare is not None:
                s = self.aftercare(s, ts.tsid, ts.pfid, ts.name)
            # Sort by dimensionality.
            dim = s.pint.dimensionality
            series_by_dimension[dim].append(s)
        # Sum to get one series per dimension.
        data = [sum(s) for s in series_by_dimension.values()]
        # Turn into portfolio line at wanted frequency.
        pfl = pf.PfLine(data)
        return pfl.asfreq(self.structure.freq)

    def general_pfl(
        self,
        pfid: str,
        tsname: str,
        ts_left: pd.Timestamp,
        ts_right: pd.Timestamp,
        missing2zero: bool = True,
        debug: bool = False,
    ) -> pf.PfLine:
        """Retrieve a portfolio line with portfolio-specific volume and/or price data
        from Belvis, without using the structure Use if wanted timeseries is not specified in the structure file.

        Parameters
        ----------
        pfid : str
            Id of portfolio as found in Belvis. Must be original.
        tsname : str
            Name of the timeseries. Must be exact.
        ts_left : pd.Timestamp
            Start of delivery period (incl)
        ts_right : pd.Timestamp
            End of delivery period (excl)
        missing2zero : bool, optional (default: True)
            What to do with values that are flagged as 'missing'. True to replace with 0,
            False to replace with nan.
        debug : bool, optional (default: False)
            If True, stops after fetching timeseries data from api; before applying the
            aftercare functions.

        Returns
        -------
        pf.PfLine
        """
        # Get ts tree and fetch data.
        ts_tree = Ts(pfid, tsname)
        self._fetch_series(ts_tree, ts_left, ts_right, missing2zero=missing2zero)
        if debug:
            return ts_tree
        # Turn into portfolio line.
        return self._pfline(ts_tree)

    def portfolio_pfl(
        self,
        pfid: str,
        pflineid: str,
        ts_left: pd.Timestamp,
        ts_right: pd.Timestamp,
        missing2zero: bool = True,
        *,
        debug: bool = False,
    ) -> pf.PfLine:
        """Retrieve a portfolio line with portfolio-specific volume and/or price data
        from Belvis.

        Parameters
        ----------
        pfid : str
            Id of portfolio as defined in .Structure. May be original or synthetic.
        pflineid : str
            Id of portfolio line as defined in .structure
        ts_left : pd.Timestamp
            Start of delivery period (incl)
        ts_right : pd.Timestamp
            End of delivery period (excl)
        missing2zero : bool, optional (default: True)
            What to do with values that are flagged as 'missing'. True to replace with 0,
            False to replace with nan.
        debug : bool, optional (default: False)
            If True, stops after fetching timeseries data from api; before applying the
            aftercare functions.

        Returns
        -------
        pf.PfLine
        """
        # Get ts trees and fetch data.
        ts_trees = []
        for pfid in self.structure.to_original_pfids(pfid):
            ts_tree = self.structure.tstree_pfline(pfid, pflineid)
            self._fetch_series(ts_tree, ts_left, ts_right, missing2zero=missing2zero)
            ts_trees.append(ts_tree)
        if debug:
            return ts_trees
        # Turn into portfolio line.
        pflines = []
        for ts_tree in ts_trees:
            pflines.append(self._pfline(ts_tree))
        return sum(pflines)

    def price_pfl(
        self,
        priceid: str,
        ts_left: pd.Timestamp,
        ts_right: pd.Timestamp,
        missing2zero: bool = True,
        *,
        debug: bool = False,
    ) -> pf.PfLine:
        """Retrieve a portfolio line with portfolio-UNspecific price data from Belvis.

        Parameters
        ----------
        priceid : str
            Id of price timeseries as defined in .structure
        ts_left : pd.Timestamp
            Start of delivery period (incl)
        ts_right : pd.Timestamp
            End of delivery period (excl)
        missing2zero : bool, optional (default: True)
            What to do with values that are flagged as 'missing'. True to replace with 0,
            False to replace with nan.

        Returns
        -------
        pf.PfLine
        """
        # Get ts trees and fetch data.
        ts_tree = self.structure.tstree_price(priceid)
        self._fetch_series(ts_tree, ts_left, ts_right, missing2zero=missing2zero)
        if debug:
            return ts_tree
        # Turn into portfolio line.
        return self._pfline(ts_tree)



from setuptools import setup
from setuptools import find_packages
import versioneer

setup(
    name="belvys",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Ruud Wijtvliet",
    zip_safe=False,
    packages=find_packages(exclude=["tests"]),
    description="Getting timeseries data from Belvis Rest API.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    python_requires=">=3.8",
    install_requires=[line.strip() for line in open("requirements.txt", "r")],
    # package_data is data that is deployed within the python package on the
    # user's system. setuptools will get whatever is listed in MANIFEST.in
    include_package_data=True,
)



"""Collection of adjustment functions."""


import datetime as dt
from typing import Callable

import pandas as pd
import portfolyo as pf

Adjustment = Callable[[pd.Series], pd.Series]
Aftercare = Callable[[pd.Series, int, str, str], pd.Series]


def fact_convert_to_tz(tz: str) -> Adjustment:
    """Returns adjustment function that converts all timestamps into timezone ``tz``."""

    def adjustment(s: pd.Series) -> pd.Series:
        return s.tz_convert(tz)

    return adjustment


convert_to_berlin: Adjustment = fact_convert_to_tz("Europe/Berlin")


def fact_fixed_to_float(fixed: str) -> Adjustment:
    """
    Use if: api returns series with frequency in incorrect (fixed) timezone.

    Parameters
    ----------
    fixed : str
        The timezone (UTC-offset) that the timestamps must be interpreted in.

    Returns
    -------
    Adjustment function

    Example
    -------
    The following timeseries has UTC timestamps but denote the start of a day in the CET
    timezone:

    2022-08-16 23:00:00+00:00      226.047
    2022-08-17 23:00:00+00:00      226.068

    The function ``ac = fact_fixed_to_float("+01:00")`` turns it into

    2022-08-17 00:00:00      226.047
    2022-08-18 00:00:00      226.068
    """

    def adjustment(s: pd.Series) -> pd.Series:
        return s.tz_convert(fixed).tz_localize(None)

    return adjustment


def fact_fixed_to_correct(fixed: str, correct: str) -> Adjustment:
    """
    Use if: api returns series with frequency in incorrect (fixed) timezone.

    Parameters
    ----------
    fixed : str
        The timezone (UTC-offset) that the timestamps must be interpreted in.
    correct : str
        The timezone that the timestamps must be converted to.

    Returns
    -------
    Adjustment function

    Example
    -------
    The following timeseries has UTC timestamps but denote the start of a day in the CET
    timezone:

    2022-08-16 23:00:00+00:00      226.047
    2022-08-17 23:00:00+00:00      226.068

    The function ``ac = fact_fixed_to_correct("+01:00", "Europe/Berlin")`` turns it into

    2022-08-17 00:00:00+02:00      226.047
    2022-08-18 00:00:00+02:00      226.068
    """

    def adjustment(s: pd.Series) -> pd.Series:
        return (
            s.tz_convert(fixed)
            .tz_localize(None)
            .tz_localize(correct, ambiguous="infer")
        )

    return adjustment


cet_to_float: Adjustment = fact_fixed_to_float("+01:00")
cet_to_berlin: Adjustment = fact_fixed_to_correct("+01:00", "Europe/Berlin")


def fact_frequency(freq: str) -> Adjustment:
    """Returns an adjustment function that infers the frequency and sets it to ``freq`` if none
    can be inferred."""

    def adjustment(s: pd.Series) -> pd.Series:
        return pf.tools.freq.set_to_frame(s, freq)

    return adjustment


hourly_frequency: Adjustment = fact_frequency("H")
quarterhourly_frequency: Adjustment = fact_frequency("15T")
daily_frequency: Adjustment = fact_frequency("D")
infer_frequency: Adjustment = fact_frequency(None)


def makeleft(s: pd.Series) -> pd.Series:
    """Assume rightbound (and turn into into leftbound) if series have <= 2h gap between
    first two timestamps."""
    td = s.index[1] - s.index[0]
    if td <= dt.timedelta(hours=2):
        s.index = pf.tools.righttoleft.index(s.index)
    return s


# def standardize_belvis(tz) -> Adjustment:
#     def adjustment(s: pd.Series, *args) -> pd.Series:
#         td = s.index[1] - s.index[0]
#         if td <= dt.timedelta(hours=2):
#             bound = "right"
#         else:
#             bound = "left"
#         return pf.standardize(s, "aware", bound, tz=tz, floating=False)

#     return adjustment
